/*
 * This file is part of libportable.
 * Copyright (C) 2008-2009 David Sveningsson <ext@sidvind.com>
 * 
 * libportable is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * libportable is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with libportable.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <portable/scandir.h>
#include <portable/asprintf.h>
#include <stdio.h>

#ifdef WIN32
#	include "_windows.h"
#endif
#include <tchar.h>

/* convert to LPCTSTR, release memory with free */
static TCHAR* to_tchar(const char* src){
	size_t len = 0;
	TCHAR* buf = NULL;

	if ( mbstowcs_s(&len, NULL, 0, src, _TRUNCATE) != 0 ){
		return NULL;
	}

	buf = (TCHAR*)malloc(sizeof(TCHAR) * (len+1));

	if ( mbstowcs_s(&len, buf, len+1, src, _TRUNCATE) != 0 ){
		return NULL;
	}

	return buf;
}

static int wildcmp(const char *wild, const char *string) {
	// Written by Jack Handy - jakkhandy@hotmail.com
	// http://www.codeproject.com/KB/string/wildcmp.aspx

	const char *cp = NULL, *mp = NULL;

	while ((*string) && (*wild != '*')) {
		if ((*wild != *string) && (*wild != '?')) {
			return 0;
		}
		wild++;
		string++;
	}

	while (*string) {
		if (*wild == '*') {
			if (!*++wild) {
				return 1;
			}
			mp = wild;
			cp = string+1;
		} else if ((*wild == *string) || (*wild == '?')) {
			wild++;
			string++;
		} else {
			wild = mp;
			string = cp++;
		}
	}

	while (*wild == '*') {
		wild++;
	}

	return !*wild;
}

int fnmatch(const char *pattern, const char *string, int flags){
	int result;

	if ( flags != 0 ){
		fprintf(stderr, "warning: fnmatch does not yet support flags, the behaviour might be unexpected.\n");
	}

	result = wildcmp(pattern, string);

	if ( result ){
		return 0;
	} else {
		return FNM_NOMATCH;
	}
}

int scandir(const char *dirp, struct dirent ***namelist,
	int (*filter)(const struct dirent *),
	int (*compar)(const struct dirent **, const struct dirent **)){

	char* dir2 = asprintf2("%s\\*", dirp);
	LPTSTR tdirp = to_tchar(dir2);
	WIN32_FIND_DATA cur;
	HANDLE handle;
	int n = 0;
	size_t list_size = 0;

	if ( compar ){
		fprintf(stderr, "scandir comparator is set, but not yet supported. No sorting will take place.\n");
	}

	handle = FindFirstFile(tdirp, &cur);
	if ( handle == INVALID_HANDLE_VALUE){
		return -1;
	}

	*namelist = NULL;

	do {
		struct dirent* d = (struct dirent*)malloc(sizeof(struct dirent));
		wcstombs_s(NULL, d->d_name, DIRENT_NAME_MAX, cur.cFileName, _TRUNCATE);

		if ( filter && !filter(d)){
			free(d);
			continue;
		}

		if ( n == list_size ){
			list_size += 25;
			*namelist = (struct dirent**)realloc(*namelist, sizeof(struct dirent*) * list_size);
		}

		(*namelist)[n++] = d;
	} while ( FindNextFile(handle, &cur) );

	free(dir2);
	free(tdirp);
	return n;
}
